//-< SortedCollection.cs >-------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// C# Program Interface                                              *  /  \  *
//                          Created:     14-Feb-2004  K.A. Knizhnik  * / [] \ *
//                          Last update: 14-Feb-2004  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Sorted collections implementation
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
    using System.Collections;

    /// <summary>
    /// Interface of sorted collection.
    /// Sorted collections keeps in members in order specified by comparator.
    /// Members in the collections can be located using key or range of keys.
    /// The SortedCollection is efficient container of objects for in-memory databases.
    /// For databases which size is significatly larger than size of page pool, operation with SortedList
    /// can cause trashing and so very bad performance. Unlike other index structures SortedCollection
    /// doesn't store values of keys and so search in the collection requires fetching of its members.
    /// </summary>
    [Metaobject(typeof(PessimisticRepeatableReadMetaobject))]
    public class SortedCollection:Persistent 
    {
        readonly PersistentComparator comparator;
        readonly bool                 unique;
        TtreePage                     root;
        int                           nMembers;
    
        protected SortedCollection(Metaobject mop) : base(mop) {}

        /// <summary>
        /// Get number of objects in the collection
        /// </summary>
        /// <returns> number of objects in the collection</returns>
        ///
        public int Length 
        { 
            [Constant]
            get 
            {
                return nMembers;
            }
        }

        /// <summary>
        /// Get number of objects in the collection
        /// </summary>
        /// <returns> number of objects in the collection</returns>
        ///
        public int Count 
        { 
            [Constant]
            get 
            {
                return nMembers;
            }
        }

        public bool IsSynchronized 
        {
            [Constant]
            get 
            {
                return true;
            }
        }

        public object SyncRoot 
        {
            [Constant]
            get 
            {
                return this;
            }
        }

        [Constant]
        public void CopyTo(Array dst, int i) 
        {
            foreach (object o in this) 
            { 
                dst.SetValue(o, i++);
            }
        }
        
        public SortedCollection(PersistentComparator comparator, bool unique) 
        { 
            this.comparator = comparator;
            this.unique = unique;
        }

        /// <summary>
        /// Get comparator used in this collection
        /// </summary>
        /// <returns> collection comparator</returns>
        ///
        [Constant]
        public PersistentComparator GetComparator() 
        { 
            return comparator;
        }

        /// <summary> Access element by key
        /// </summary>
        public Persistent this[object key] 
        { 
            [Constant]
            get 
            {
                if (root != null) 
                  { 
                      ArrayList list = new ArrayList();
                      root.find(comparator, key, key, list);
                      if (list.Count > 1) 
                      { 
                          throw new System.InvalidOperationException("key not unique");
                      } 
                      else if (list.Count == 0) 
                      { 
                          return null;
                      } 
                      else 
                      { 
                          return (Persistent)list[0];
                      }
                  }
                return null;
            }
        }

        /// <summary> Access elements by key range
        /// </summary>
         public Persistent[] this[object low, object high] 
        { 
            [Constant]
            get 
            { 
                ArrayList list = new ArrayList();
                if (root != null) 
                { 
                
                    root.find(comparator, low, high, list);
                }
                return (Persistent[])list.ToArray(typeof(Persistent));
            }
        }


        /// <summary>
        /// Add new member to collection
        /// </summary>
        /// <param name="obj"> new member</param>
        /// <returns> <code>true</code> if object is successfully added in the index, 
        /// <code>false</code> if collection was declared as unique and there is already member with such value
        /// of the key in the collection. </returns>
        ///
        public bool Add(Persistent obj) 
        { 
            TtreePage newRoot = root;
            if (root == null) 
            { 
                newRoot = new TtreePage(obj);
            } 
            else 
            { 
                if (root.insert(comparator, obj, unique, ref newRoot) == TtreePage.NOT_UNIQUE) 
                { 
                    return false;
                }
            }
            Modify();
            root = newRoot;
            nMembers += 1;
            return true;
        }
                
                
        /// <summary>
        /// Check if collections contains specified member
        /// </summary>
        /// <returns> <code>true</code> if specified member belongs to the collection</returns>
        ///
        [Constant]
        public bool Contains(Persistent member) 
        {
            return (root != null) ? root.contains(comparator, member) : false;
        }        

        /// <summary>
        /// Remove member from collection
        /// </summary>
        /// <param name="obj"> member to be removed</param>
        /// <returns><code>true</code> if object is deleted, <code>false</code> if objec tnot found</returns>
        ///
        public bool Remove(Persistent obj) 
        {
            if (root == null) 
            {
                return false;
            }
            TtreePage newRoot = root;
            if (root.remove(comparator, obj, ref newRoot) == TtreePage.NOT_FOUND) 
            {             
                return false;
            }
            Modify();
            root = newRoot;
            nMembers -= 1;  
            return true;
        }

    
        /// <summary>
        /// Remove all objects from the collection
        /// </summary>
        ///
        public void Clear() 
        {
            if (root != null) 
            { 
                root.prune();
                Modify();
                root = null;
                nMembers = 0;
            }
        }
 
        /// <summary>
        /// Get all objects in the index as array ordered by index key.
        /// </summary>
        /// <returns> array of objects in the index ordered by key value</returns>
        ///
        [Constant]
        public Persistent[] ToArray() 
        {
            Persistent[] arr = new Persistent[nMembers];
            if (root != null) 
            { 
                root.toArray(arr, 0);
            }
            return arr;
        }

        /// <summary> Get all objects in the index as array of specified type orderd by index key
        /// </summary>
        /// <param name="elemType">type of array element</param>
        /// <returns>array of objects in the index ordered by key value
        /// </returns>
        [Constant]
        public virtual Array ToArray(Type elemType)
        {
            Array arr = Array.CreateInstance(elemType, nMembers);
            if (root != null)
            {
                root.toArray((Persistent[])arr, 0);
            }
            return arr;
        }

 
        class TtreeEnumerator : IEnumerator 
        { 
            int              i;
            ArrayList        list;
            SortedCollection tree;

            internal TtreeEnumerator(SortedCollection tree, ArrayList list) 
            { 
                this.tree = tree;
                this.list = list;
                i = -1;
            }
        
            public void Reset() 
            {
                i = -1;
            }
                
            public object Current
            {
                get 
                {
                    if (i < 0 || i >= list.Count) 
                    { 
                        throw new InvalidOperationException();
                    }
                    return list[i];
                }
            }

            public bool MoveNext() 
            {
                if (i+1 < list.Count) 
                { 
                    i += 1;
                    return true;
                }
                return false;
            }
        }
        

        /// <summary>
        /// Get iterator for traversing collection members  with key belonging to the specified range. 
        /// </summary>
        /// <returns> selection iterator</returns>
        ///
        [Constant]
        public IEnumerator GetEnumerator()
        {
            return GetEnumerator(null, null);
        }

        /// <summary>
        /// Get iterator for traversing collection members  with key belonging to the specified range. 
        /// </summary>
        /// <param name="from"> low boundary. If <code>null</code> then low boundary is not specified.
        /// Low boundary can be inclusive or exclusive.</param>
        /// <param name="till"> high boundary. If <code>null</code> then high boundary is not specified.
        /// High boundary can be inclusive or exclusive. </param>
        /// <returns> selection iterator</returns>
        ///
        [Constant]
        public IEnumerator GetEnumerator(object from, object till) 
        {
            ArrayList list = new ArrayList();
            if (root != null) 
            { 
                root.find(comparator, from, till, list);
            }            
            return new TtreeEnumerator(this, list);
        }
    }
}
