//-< TESTKIT.CXX >---------------------------------------------------*--------*
//-------------------------------------------------------------------*--------*
// Example of using kit iterators to perform linear database searches
//-------------------------------------------------------------------*--------*

#include "goods.h"
#include "kit.h"

USE_GOODS_NAMESPACE


class ObjA : public object
{
public:
   ObjA() : object(self_class)  {}

   // define STL like interface to iterators
   typedef const_iterator_ref<ObjA> ro_iterator;
   typedef iterator_ref<ObjA>       rw_iterator;

   // this method is required by iterators (if it isn't define you got a compile time error)
   inline const ref<ObjA> getNext() const // for read-only iterators
      { return next; }

   inline ref<ObjA> getNext() // for read-write iterators
      { return next; }

   ref<ObjA> next;

   // some data fields
   nat4 x;
   int4 y;

   METACLASS_DECLARATIONS(ObjA, object);
};


class ObjB : public object
{
public:
   ObjB() : object(self_class)  {}

   // define STL like interface to iterators
   typedef const_iterator_ref<ObjB> ro_iterator;
   typedef iterator_ref<ObjB>       rw_iterator;

   // this method is required by iterators (if it isn't define you got a compile time error)
   inline const ref<ObjB> getNext() const // for read-only iterators
      { return next; }

   inline ref<ObjB> getNext() // for read-write iterators
      { return next; }

   ref<ObjB> next;

   // some data fields
   ref<String> s1;
   ref<String> s2;

   METACLASS_DECLARATIONS(ObjB, object);
};


class CRoot : public object
{
public:
   CRoot() : object(self_class)  {}

   void initialize() const
   { 
      if (is_abstract_root()) 
      {
         ref<CRoot> root = this;
         modify(root)->become(new CRoot);
      } 
   }

   void addObjA(ObjA* pA)
   {
      pA->next = firstA;
      firstA = pA;
   }
   void addObjB(ObjB* pB)
   {
      pB->next = firstB;
      firstB = pB;
   }

   ref<ObjA> firstA;
   ref<ObjB> firstB;

   METACLASS_DECLARATIONS(CRoot, object);
};


int main() 
{
   database db;
   task::initialize(task::huge_stack);
   if(db.open("testdb.cfg"))
   { 
      ref<CRoot> root;
      db.get_root(root);
      root->initialize();

      // ---------------------------------------------------------------------------------- //
      
      modify(root)->firstA = NULL;
      modify(root)->firstB = NULL;

      // add some stuff...
      ObjA *pA = new ObjA;
      pA->x = 5;
      pA->y = 2;
      modify(root)->addObjA(pA);

      pA = new ObjA;
      pA->x = 10;
      pA->y = -9;
      modify(root)->addObjA(pA);

      pA = new ObjA;
      pA->x = 11;
      pA->y = 2;
      modify(root)->addObjA(pA);

      // ...
      ObjB *pB = new ObjB;
      pB->s1 = String::create("test");
      pB->s2 = String::create("TesT");
      modify(root)->addObjB(pB);

      pB = new ObjB;
      pB->s1 = String::create("yes");
      pB->s2 = String::create("no");
      modify(root)->addObjB(pB);

      pB = new ObjB;
      pB->s1 = String::create("test");
      pB->s2 = String::create("no");
      modify(root)->addObjB(pB);

      // ---------------------------------------------------------------------------------- //

      // let's do some iterator work
      {
         printf("iterating on all ObjA elements\n");
         ObjA::ro_iterator it(root->firstA);
         while(it)
         {
            printf("A(%u, %d)\n", it->x, it->y);
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjA elements with y = 2\n");
         ObjA::ro_iterator it(root->firstA, kit::qequal<ObjA, int4>(&ObjA::y, 2));
         while(it)
         {
            printf("A(%u, %d)\n", it->x, it->y);
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjA elements with y > 0 and x != 11\n");
         kit::query<ObjA> q = kit::qgreat<ObjA, int4>(&ObjA::y, 0) && !kit::qequal<ObjA, nat4>(&ObjA::x, 11);
         ObjA::ro_iterator it(root->firstA, q);
         while(it)
         {
            printf("A(%u, %d)\n", it->x, it->y);
            ++it;
         }
         printf("\n");
      }

      // iterate on ObjB
      {
         printf("iterating on all ObjB elements\n");
         ObjB::ro_iterator it(root->firstB);
         while(it)
         {
            printf("B(%s, %s)\n", it->s1->get_text(), it->s2->get_text());
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjB elements with s1 = test\n");
         ObjB::ro_iterator it(root->firstB, kit::qequal<ObjB, ref<String>, cmpString>(&ObjB::s1, String::create("test")));
         while(it)
         {
            printf("B(%s, %s)\n", it->s1->get_text(), it->s2->get_text());
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjB elements with s1 = Test (case insensitive)\n");
         ObjB::ro_iterator it(root->firstB, kit::qequal<ObjB, ref<String>, cmpIString>(&ObjB::s1, String::create("Test")));
         while(it)
         {
            printf("B(%s, %s)\n", it->s1->get_text(), it->s2->get_text());
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjB elements with s1 = Test (case insensitive)\n");
         ObjB::ro_iterator it(root->firstB, kit::qequal<ObjB, ref<String>, cmpIString>(&ObjB::s1, String::create("Test")));
         while(it)
         {
            printf("B(%s, %s)\n", it->s1->get_text(), it->s2->get_text());
            ++it;
         }
         printf("\n");
      }
      {
         printf("iterating on all ObjB elements with s1 = yes or (s2 != no and s1 == test)\n");

         kit::qequal<ObjB, ref<String>, cmpIString> eqYes(&ObjB::s1, String::create("yes"));
         kit::qequal<ObjB, ref<String>, cmpIString> eqNo(&ObjB::s2, String::create("no"));
         kit::qequal<ObjB, ref<String>, cmpIString> eqTest(&ObjB::s1, String::create("test"));

         ObjB::ro_iterator it(root->firstB, eqYes || (!eqNo && eqTest));
         while(it)
         {
            printf("B(%s, %s)\n", it->s1->get_text(), it->s2->get_text());
            ++it;
         }
         printf("\n");
      }

      // NOTE: you can combine queries any way you want and how many you want with standard operators.

      // example of read write iterators
      {
         printf("setting x = 4 to all ObjA with y = 2\n");

         ObjA::rw_iterator it(root->firstA, kit::qequal<ObjA, int4>(&ObjA::y, 2));
         while(it)
         {
            printf("A(%u, %d) <-- A(", it->x, it->y);
            it->x = 4;
            printf("%u, %d)\n", it->x, it->y);
            ++it;
         }
         printf("\n");
      }

      // HERE (on "it" destructor) transaction is committed
      // NOTE: you can force transaction commit calling iterator release() method;
      // NOTE: you can get object pointer calling get() iterator method

      {
         printf("iterating on all ObjA elements with x = 4\n");
         ObjA::ro_iterator it(root->firstA, kit::qequal<ObjA, nat4>(&ObjA::x, 4));
         while(it)
         {
            printf("A(%u, %d)\n", it->x, it->y);
            ++it;
         }
         printf("\n");
      }

   }
   else 
   { 
      fprintf(stderr, "Failed to open database\n");
      return 1;
   }
}


field_descriptor& ObjA::describe_components()
{
    return FIELD(next), FIELD(x), FIELD(y); 
}

field_descriptor& ObjB::describe_components()
{
    return FIELD(next), FIELD(s1), FIELD(s2); 
}

field_descriptor& CRoot::describe_components()
{
    return FIELD(firstA), FIELD(firstB); 
}


REGISTER(ObjA, object, pessimistic_repeatable_read_scheme);
REGISTER(ObjB, object, pessimistic_repeatable_read_scheme);
REGISTER(CRoot, object, pessimistic_repeatable_read_scheme);
