//-< TESTTRANS.CXX >-------------------------------------------------*--------*
// GOODS                     Version 2.37        (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     15-Jan-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 15-Jan-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test of per-thread transactions
//-------------------------------------------------------------------*--------*

#include "goods.h"

USE_GOODS_NAMESPACE

const int nNodes = 1000;
const int nIterations = 1000;
const int nThreadsByDefault = 10;

static database db;
static semaphore done;

class L2ListNode: public object { 
  public:
    METACLASS_DECLARATIONS(L2ListNode, object);

    ref<L2ListNode> next;
    ref<L2ListNode> prev;
    int             value;

    L2ListNode(class_descriptor& desc = self_class) : object(desc) {}

    void increment() { 
        value += 1;
    }

    void decrement() { 
        value -= 1; 
    }

    void update() { 
        increment();
        modify(next)->decrement();
    }
};

field_descriptor& L2ListNode::describe_components()
{
    return FIELD(next), FIELD(prev), FIELD(value);
}


REGISTER(L2ListNode, object, hierarchical_access_scheme);

class L2List: public L2ListNode { 
    int n_members;
  public:
    L2List() : L2ListNode(self_class) {}

    int size() const { 
        return n_members;
    }

    void link(L2ListNode* node) {
        node->next = this;
        node->prev = prev;
        modify(prev)->next = node;
        prev = node;
        n_members += 1;
    }


    void buildList() { 
        if (next.is_nil()) { 
            next = this;
            prev = this;
            value = 0;
            n_members = 0;
            for (int i = 1; i <= nNodes; i++) {
                L2ListNode* node = new L2ListNode();
                node->value = i;
                link(node);
            }
        }
    }

    int calculateSum() const { 
        int sum = 0;
        for (ref<L2ListNode> node = next; node != this; node = node->next) {
            sum += node->value;
        }
        return sum;
    }
    
    void updateNode(int index) const { 
        ref<L2ListNode> node = next; 
        while (--index >= 0) { 
            node = node->next;
        }
        modify(node)->update();
    }   

    void reset() { 
        next = NULL;
        prev = NULL;
    }

    void initialize() const { 
        if (is_abstract_root()) { 
            ref<L2List> root = this;
            modify(root)->become(new L2List());
        }
    }

    METACLASS_DECLARATIONS(L2List, L2ListNode);
};

field_descriptor& L2List::describe_components()
{
    return FIELD(n_members);
}

REGISTER(L2List, L2ListNode, pessimistic_concurrent_repeatable_read_scheme);


void task_proc run(void*) { 
    db.attach();
    ref<L2List> list;
    db.get_root(list);
    update(list)->initialize();
    modify(list)->buildList();

    for (int i = 0; i < nIterations; i++) { 
        int sum = list->calculateSum();
        if (sum != nNodes*(nNodes+1)/2) { 
            fprintf(stderr, "Wrong check sum: %d instead of %d\n",   
                    sum, nNodes*(nNodes+1)/2);
        }
        update(list)->updateNode(i % (nNodes-1));
    }
    db.detach();
    done.signal();
}
 
int main(int argc, char* argv[])
{
    int nThreads = nThreadsByDefault;
    int i;
    task::initialize(task::normal_stack);
    if (argc > 1) { 
        nThreads = atoi(argv[1]);
    }
    transaction_manager::set_isolation_level(PER_THREAD_TRANSACTION);
    if (db.open("testtrans.cfg")) {
        time_t start = time(NULL);
        for (i = 0; i < nThreads; i++) { 
            task::create(run);
        }
        for (i = 0; i < nThreads; i++) { 
            done.wait();
        }
        printf("Elapsed time %ld\n", long(time(NULL) - start));
        ref<L2List> list;
        db.get_root(list);
        if (list->size() != nNodes
            || list->next->value != 1 + (nIterations+nNodes-2)/(nNodes-1)*nThreads 
            || list->prev->value != nNodes - nIterations/(nNodes-1)*nThreads) 
        { 
            fprintf(stderr, "Something is definitly wrong: list->next->value=%d instead of %d, "
                    "list->prev->value=%d instead of %d\n", 
                    list->next->value, 1 + (nIterations+nNodes-2)/(nNodes-1)*nThreads,  
                    list->prev->value, nNodes - nIterations/(nNodes-1)*nThreads);
        } else { 
            printf("Test successfully passed\n");
        }       
        modify(list)->reset();
        db.close();
        return EXIT_SUCCESS;
    } else {
        fprintf(stderr, "Failed to open detabase\n");
        return EXIT_FAILURE;
    }
}





