//-< Rtree.java >----------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// GOODS Persistent Class Library                                    *  /  \  *
//                          Created:     17-Mar-2005  K.A. Knizhnik  * / [] \ *
//                          Last update: 17-Mar-2005  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// R-Tree: spatial index
//-------------------------------------------------------------------*--------*

package goodslib;
import  goodsjpi.*;
import java.util.*;

public class Rtree extends Persistent {
    private int           height;
    private int           n;
    private RtreePage     root;
    private transient int updateCounter;

    Rtree() {}

    /**
     * Put new spatial object in the index. 
     * @param obj spatial object with specified bounded rectangle 
     */
    public void put(SpatialObject obj) {
        if (root == null) { 
            root = new RtreePage(obj);
            height = 1;
        } else { 
            RtreePage p = root.insert(obj, height); 
            if (p != null) {
                root = new RtreePage(root, p);
                height += 1;
            }
        }
        n += 1;
        updateCounter += 1;
    }
    
    /**
     * Get number of objects in the index
     * @return number of objects in the index
     */
    public int size() { 
        return n;
    }

    /**
     * Remove spatial object from the tree.
     * @param obj spatial object with specified bounded rectangle 
     * @return <code>true</code> if object was sucessfully deleted, <code>false</code> oif
     * object not found in tree
     */
    public boolean remove(SpatialObject obj) {
        if (root == null) { 
            return false;
        }
        ArrayList reinsertList = new ArrayList();
        int reinsertLevel = root.remove(obj, height, reinsertList);
        if (reinsertLevel < 0) { 
            return false;
        }        
        for (int i = reinsertList.size(); --i >= 0;) {
            RtreePage p = (RtreePage)reinsertList.get(i);
            for (int j = 0, n = p.n; j < n; j++) { 
                RtreePage q = root.insert(p.branch[j], height - reinsertLevel); 
                if (q != null) { 
                    // root splitted
                    root = new RtreePage(root, q);
                    height += 1;
                }
            }
            reinsertLevel -= 1;
        }
        if (root.n == 1 && height > 1) { 
            RtreePage newRoot = (RtreePage)root.branch[0];
            root = newRoot;
            height -= 1;
        }
        n -= 1;
        updateCounter += 1;
        return true;
    }
    
    /**
     * Find all objects located in the selected rectangle
     * @param r selected rectangle
     * @return array of objects which enveloping rectangle intersects with specified rectangle
     */
    public SpatialObject[] get(Rectangle r) {
        ArrayList result = new ArrayList();
        if (root != null) { 
            root.find(r, result, height);
        }
        return (SpatialObject[])result.toArray(new SpatialObject[result.size()]);
    }

    /**
     * Find all objects located in the selected rectangle
     * @param r selected rectangle
     * @return array list of objects which enveloping rectangle intersects with specified rectangle
     */
    public ArrayList getList(Rectangle r) { 
        ArrayList result = new ArrayList();
        if (root != null) { 
            root.find(r, result, height);
        }
        return result;
    }

    /**
     * Get array of all members of the index
     * @return array of index members
     */
    public SpatialObject[] toArray() {
        return get(getWrappingRectangle());
    }

    /**
     * Get all objects in the index.
     * The runtime type of the returned array is that of the specified array.  
     * If the index fits in the specified array, it is returned therein.  
     * Otherwise, a new array is allocated with the runtime type of the 
     * specified array and the size of this index.<p>
     *
     * If this index fits in the specified array with room to spare
     * (i.e., the array has more elements than this index), the element
     * in the array immediately following the end of the index is set to
     * <tt>null</tt>.  This is useful in determining the length of this
     * index <i>only</i> if the caller knows that this index does
     * not contain any <tt>null</tt> elements.)<p>
     * @param arr specified array
     * @return array of all objects in the index
     */
    public SpatialObject[] toArray(SpatialObject[] arr) {
        return (SpatialObject[])getList(getWrappingRectangle()).toArray(arr);
    }

    /**
     * Get wrapping rectangle 
     * @return minimal rectangle containing all rectangles in the index, <code>null</code> if index is empty     
     */
    public Rectangle getWrappingRectangle() {
        if (root != null) { 
            return root.getRectangle();
        }
        return null;
    }

    /**
     * Remove all objects from the index
     */
    public void clear() {
        if (root != null) { 
            root.purge(height);
            root = null;
        }
        height = 0;
        n = 0;
        updateCounter += 1;
    }

    class RtreeIterator implements Iterator {
        RtreeIterator(Rectangle r) { 
            counter = updateCounter;
            if (height == 0) { 
                return;
            }
            this.r = r;            
            pageStack = new RtreePage[height];
            posStack = new int[height];

            if (!gotoFirstItem(0, root)) { 
                pageStack = null;
                posStack = null;
            }
        }

        public boolean hasNext() {
            if (counter != updateCounter) { 
                throw new ConcurrentModificationException();
            }
            return pageStack != null;
        }

        protected Object current(int sp) { 
            return pageStack[sp].branch[posStack[sp]];
        }

        public Object next() {
            if (!hasNext()) { 
                throw new NoSuchElementException();
            }
            Object curr = current(height-1);
            if (!gotoNextItem(height-1)) { 
                pageStack = null;
                posStack = null;
            }
            return curr;
        }
 
        private boolean gotoFirstItem(int sp, RtreePage pg) { 
            for (int i = 0, n = pg.n; i < n; i++) { 
                if (r.intersects(pg.getRectangle(i))) { 
                    if (sp+1 == height || gotoFirstItem(sp+1, (RtreePage)pg.branch[i])) { 
                        pageStack[sp] = pg;
                        posStack[sp] = i;
                        return true;
                    }
                }
            }
            return false;
        }
              
 
        private boolean gotoNextItem(int sp) {
            RtreePage pg = pageStack[sp];
            for (int i = posStack[sp], n = pg.n; ++i < n;) { 
                if (r.intersects(pg.getRectangle(i))) { 
                    if (sp+1 == height || gotoFirstItem(sp+1, (RtreePage)pg.branch[i])) { 
                        pageStack[sp] = pg;
                        posStack[sp] = i;
                        return true;
                    }
                }
            }
            pageStack[sp] = null;
            return (sp > 0) ? gotoNextItem(sp-1) : false;
        }
              
        public void remove() { 
            throw new UnsupportedOperationException();
        }

        RtreePage[] pageStack;
        int[]       posStack;
        int         counter;
        Rectangle   r;
    }
    
    /**
     * Get iterator through all members of the index
     * @return iterator through all objects in the index
     */
    public Iterator iterator() {
        return iterator(getWrappingRectangle());
    }

    /**
     * Get objects which rectangle intersects with specified rectangle
     * @param r selected rectangle
     * @return iterator for objects which enveloping rectangle overlaps with specified rectangle
     */
    public Iterator iterator(Rectangle r) { 
        return new RtreeIterator(r);
    }
}
    

